@@ -17,15 +17,21 @@
     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
 
+#include <QFileInfo>
 #include <QHash>
 #include <QList>
+#include <QLibrary>
 #include <QStack>
+#include <QDir>
 
 #include <type.h>
+#include <smoke.h>
 
 #include "globals.h"
 #include "../../options.h"
 
+typedef void (*InitSmokeFn)();
+
 QHash<QString, QString> Util::typeMap;
 QHash<const Method*, const Function*> Util::globalFunctionMap;
 QHash<const Method*, const Field*> Util::fieldAccessors;
@@ -92,13 +98,116 @@ bool operator==(const EnumMember& lhs, const EnumMember& rhs)
     return (lhs.name() == rhs.name() && lhs.declaringType() == rhs.declaringType() && lhs.type() == rhs.type());
 }
 
+static Smoke* loadSmokeModule(QString moduleName) {
+    QLibrary lib;
+#if defined(Q_OS_WIN32)
+    QString libName = QLatin1String("smoke") + moduleName;
+#else
+    QString libName = QLatin1String("libsmoke") + moduleName;
+#endif
+
+    // first, try <libdir>/moduleName/libsmokemoduleName
+    lib.setFileName(Options::libDir.filePath(moduleName + '/' + libName));
+
+    // then <libdir>/libsmokemoduleName
+    if (!lib.load()) {
+        lib.setFileName(Options::libDir.filePath(libName));
+    }
+
+    // use the plain library name if everything else fails
+    if (!lib.load()) {
+        lib.setFileName(libName);
+    }
+
+    lib.load();
+
+    QString init_name = "init_" + moduleName + "_Smoke";
+    InitSmokeFn init = (InitSmokeFn) lib.resolve(init_name.toLatin1());
+
+    if (!init) {
+        qWarning("Couldn't resolve %s: %s", qPrintable(init_name), qPrintable(lib.errorString()));
+        return 0;
+    }
+
+    (*init)();
+
+    QString smoke_name = moduleName + "_Smoke";
+    Smoke** smoke = (Smoke**) lib.resolve(smoke_name.toLatin1());
+    if (!smoke) {
+        qWarning("Couldn't resolve %s: %s", qPrintable(smoke_name), qPrintable(lib.errorString()));
+        return 0;
+    }
+
+    return *smoke;
+}
+
+static bool compareArgs(const Method& method, const Smoke::Method& smokeMethod, Smoke* smoke) {
+    if (method.parameters().count() != smokeMethod.numArgs) {
+        return false;
+    }
+    for (int i = 0; i < method.parameters().count(); i++) {
+        const Parameter& p = method.parameters()[i];
+        if (p.type()->toString() != QLatin1String(smoke->types[smoke->argumentList[smokeMethod.args + i]].name)) {
+            return false;
+        }
+    }
+    return true;
+}
+
+static bool isRepeating(const QList<Smoke*>& parentModules, const char* className, const Method& method) {
+    QString mungedName = Util::mungedName(method).toLatin1();
+    foreach (Smoke* smoke, parentModules) {
+        Smoke::ModuleIndex methodIndex = smoke->findMethod(className, mungedName.toLatin1().constData());
+        if (methodIndex.index) {
+            Smoke::Index index = methodIndex.smoke->methodMaps[methodIndex.index].method;
+            if (index >= 0) {
+                if (compareArgs(method, methodIndex.smoke->methods[index], methodIndex.smoke)) {
+                    return true;
+                }
+                continue;
+            }
+            index = -index;
+            Smoke::Index i;
+            while ((i = methodIndex.smoke->ambiguousMethodList[index++]) != 0) {
+                if (compareArgs(method, methodIndex.smoke->methods[i], methodIndex.smoke)) {
+                    return true;
+                }
+            }
+        }
+    }
+    return false;
+}
+
+// assuming that enums don't change between modules, checking for the first member only is sufficient
+static bool isRepeating(const QList<Smoke*>& parentModules, const char* className, const Enum& eNum) {
+    if (eNum.members().isEmpty())
+        return false;
+
+    const EnumMember& firstMember = eNum.members().first();
+
+    foreach(Smoke *smoke, parentModules) {
+        Smoke::ModuleIndex methodIndex = smoke->findMethod(className, firstMember.name().toLatin1().constData());
+        if (methodIndex.index)
+            return true;
+    }
+    return false;
+}
+
 void Util::preparse(QSet<Type*> *usedTypes, QSet<const Class*> *superClasses, const QList<QString>& keys)
 {
     Class& globalSpace = classes["QGlobalSpace"];
     globalSpace.setName("QGlobalSpace");
     globalSpace.setKind(Class::Kind_Class);
     globalSpace.setIsNameSpace(true);
-    
+
+    QList<Smoke*> parentModules;
+    foreach (QString module, Options::parentModules) {
+        Smoke *smoke = loadSmokeModule(module);
+        if (smoke) {
+            parentModules << smoke;
+        }
+    }
+
     // add all functions as methods to a class called 'QGlobalSpace' or a class that represents a namespace
     for (QHash<QString, Function>::const_iterator it = functions.constBegin(); it != functions.constEnd(); it++) {
         const Function& fn = it.value();
@@ -127,6 +236,9 @@ void Util::preparse(QSet<Type*> *usedTypes, QSet<const Class*> *superClasses, co
         
         Method meth = Method(parent, fn.name(), fn.type(), Access_public, fn.parameters());
         meth.setFlag(Method::Static);
+        if (isRepeating(parentModules, parent->name().toLatin1(), meth)) {
+            continue;
+        }
         parent->appendMethod(meth);
         // map this method to the function, so we can later retrieve the header it was defined in
         globalFunctionMap[&parent->methods().last()] = &fn;
@@ -141,12 +253,13 @@ void Util::preparse(QSet<Type*> *usedTypes, QSet<const Class*> *superClasses, co
         foreach (const Parameter& param, meth.parameters())
             (*usedTypes) << param.type();
     }
-    
+
     // all enums that don't have a parent are put under QGlobalSpace, too
     for (QHash<QString, Enum>::iterator it = enums.begin(); it != enums.end(); it++) {
         Enum& e = it.value();
         if (!e.parent()) {
             Class* parent = &globalSpace;
+            // if the enum is defined in a namespace, make that the enum's parent
             if (!e.nameSpace().isEmpty()) {
                 parent = &classes[e.nameSpace()];
                 if (parent->name().isEmpty()) {
@@ -154,6 +267,9 @@ void Util::preparse(QSet<Type*> *usedTypes, QSet<const Class*> *superClasses, co
                     parent->setKind(Class::Kind_Class);
                     parent->setIsNameSpace(true);
                 }
+            // else, see if it is already defined in a parent module
+            } else if (isRepeating(parentModules, parent->name().toLatin1(), e)) {
+                continue;
             }
 
             Type *t = 0;
@@ -169,7 +285,11 @@ void Util::preparse(QSet<Type*> *usedTypes, QSet<const Class*> *superClasses, co
             parent->appendChild(&e);
         }
     }
-    
+
+    foreach (Smoke* smoke, parentModules) {
+        delete smoke;
+    }
+
     foreach (const QString& key, keys) {
         Class& klass = classes[key];
         foreach (const Class::BaseClassSpecifier base, klass.baseClasses()) {
